#include "mainwindow.h"
#include "dbio.h"
#include "displayutil.h"
#include "brlcad/ConstDatabase.h"
#include "brlcad/Object.h"
#include "brlcad/Combination.h"
#include <QApplication>
#include <QMenuBar>
#include <QAction>
#include <QKeySequence>
#include <QFileDialog>
#include <QStringList>
#include <QListWidget>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QMessageBox>
#include <QImage>
#include <QLabel>
#include <QPixmap>
#include <QSplitter>
#include <QPushButton>
#include <string>
#include <iostream>


void
MainWindow::setDatabaseName(const char *databaseName)
{
	std::string title = "";
	if (databaseName) {
		title += databaseName;
		title += " - BRLCAD-GUI";
	} else {
		title += "BRLCAD-GUI";
	}
	setWindowTitle(title.c_str());
}


void
MainWindow::openFileDialog()
{
	QFileDialog *dialog = new QFileDialog();
	QStringList *filters = new QStringList();
	*filters << "BRL-CAD Database Files (*.g)"
		    << "All files (*)";
	dialog->setNameFilters(*filters);
	dialog->setFileMode(QFileDialog::AnyFile);
	if (dialog->exec()) {
    	QStringList fileNames = dialog->selectedFiles();
    	std::string fileName = fileNames[0].toStdString();
    	const char *fileNamePtr = fileName.c_str();
    	std::cout << fileNamePtr << std::endl;
    	if (database)
    	 	closeDatabase();
    	loadDatabase(fileNamePtr);
	}
}


void
MainWindow::loadDatabase(const char *fileName)
{
	if (DatabaseUtils::openDatabase(database, fileName) == 0) {
		setDatabaseName(database->Title());
		std::vector<const char*> objects = loadTopLevelObjects();
		setInterfaceVisible(true);
    }
}

void
MainWindow::closeDatabase()
{
	setDatabaseName(NULL);
	setInterfaceVisible(false);
}


void
MainWindow::setInterfaceVisible(bool visible)
{
	if (visible) {
		actionClose->setVisible(true);

	} else {
		actionClose->setVisible(false);
		objectListWidget->clear();
	}
}


void
MainWindow::quit()
{
	QApplication::instance()->quit();
}


void
MainWindow::initMenu()
{
	QMenuBar *menuBar = this->menuBar();

	actionOpen = new QAction(tr("Open Database"), this);
	actionOpen->setShortcut(QKeySequence::Open);
	connect(actionOpen, &QAction::triggered, this, &MainWindow::openFileDialog);

	actionClose = new QAction(tr("Close Database"), this);
	actionClose->setShortcut(QKeySequence::Close);
	connect(actionClose, &QAction::triggered, this, &MainWindow::closeDatabase);
	actionClose->setVisible(false);

	actionQuit = new QAction(tr("Quit"), this);
	actionQuit->setShortcut(QKeySequence::Quit);
	connect(actionQuit, &QAction::triggered, this, &MainWindow::quit);

	QPushButton *rayTraceButton = new QPushButton(this);
	rayTraceButton->setText(tr("Raytrace"));

	QMenu *fileMenu = menuBar->addMenu(tr("&File"));
	fileMenu->addAction(actionOpen);
	fileMenu->addSeparator();
	fileMenu->addAction(actionClose);
	fileMenu->addSeparator();
	fileMenu->addAction(actionQuit);

	objectListWidget = new QListWidget(this);
	connect(objectListWidget, &QListWidget::itemDoubleClicked, 
            this, &MainWindow::onObjectClicked);

	connect(rayTraceButton, &QPushButton::clicked, this, &MainWindow::rayTraceSelectedListItem);

	QLabel *objectListLabel = new QLabel(this);
	objectListLabel->setText(tr("Objects"));

	mainDisplayWidget = new QLabel(this);

	QSplitter *centralContainer = new QSplitter(this);
	QWidget *objectListContainerWrapper = new QWidget(this);
	QVBoxLayout *objectListContainer = new QVBoxLayout(objectListContainerWrapper);

	centralContainer->setOrientation(Qt::Horizontal);
	centralContainer->addWidget(objectListContainerWrapper);
	centralContainer->addWidget(mainDisplayWidget);
	centralContainer->setStretchFactor(1, 3);


	objectListContainer->addWidget(objectListLabel);
	objectListContainer->addWidget(objectListWidget);
	objectListContainer->addWidget(rayTraceButton);

	setCentralWidget(centralContainer);
}


void
MainWindow::rayTraceSelectedListItem()
{
	if (objectListWidget->currentItem()) {
		std::string objectName = objectListWidget->currentItem()->text().toStdString();
		const char *objectNamePtr = objectName.c_str();
		rayTrace(objectNamePtr);
	}
}


void
MainWindow::rayTrace(const char *object)
{
	Camera *camera = new Camera();
	QPixmap *pixmap = DisplayUtils::rayTrace(database, object, camera);
	mainDisplayWidget->setPixmap(*pixmap);
}


void
MainWindow::onObjectClicked(QListWidgetItem* listItem)
{
	std::string objectName = listItem->text().toStdString();
	const char *objectNamePtr = objectName.c_str();
	database->UnSelectAll();
	database->Select(objectNamePtr);
	BRLCAD::Vector3D boxMinima = database->BoundingBoxMinima();
	BRLCAD::Vector3D boxMaxima = database->BoundingBoxMaxima();
	const double *minimaCoordinates = boxMinima.coordinates;
	const double *maximaCoordinates = boxMaxima.coordinates;
	QString windowTitle = QString("Object \"%1\"").arg(objectNamePtr);
	QMessageBox *messageBox = new QMessageBox(this);
	QString minimaText = "BoundingBoxMinima: (%1, %2, %3)";
	QString maximaText = "BoundingBoxMaxima: (%1, %2, %3)";

	messageBox->setWindowTitle(windowTitle);
	minimaText = minimaText.arg(minimaCoordinates[0]).arg(minimaCoordinates[1]).arg(minimaCoordinates[2]);
	maximaText = maximaText.arg(maximaCoordinates[0]).arg(maximaCoordinates[1]).arg(maximaCoordinates[2]);
	messageBox->setText(minimaText+"\n"+maximaText);
	messageBox->exec();
}


std::vector<const char*>
MainWindow::loadTopLevelObjects()
{
	Camera *camera = new Camera();
	BRLCAD::Vector3D *vec = new BRLCAD::Vector3D(40, 50, 1);
	std::cout << "Test vector:" << vec->coordinates[0] << " " << vec->coordinates[1] << " " << vec->coordinates[2] << std::endl;

	double *twod = DisplayUtils::mapTo2D(vec, camera);
	std::cout << "2Ded vector:" << twod[0] << " " << twod[1] << std::endl;

	BRLCAD::Vector3D *recovered = DisplayUtils::mapTo3D(twod, camera);
	std::cout << "Recovered:" << recovered->coordinates[0] << " " << recovered->coordinates[1] << " " << recovered->coordinates[2] << std::endl;
	std::vector<const char*> objects;

	BRLCAD::ConstDatabase::TopObjectIterator iter = database->FirstTopObject();
	while (iter.Good())  {
		const char *objectName = iter.Name();

		std::cout << objectName << std::endl;
		objects.push_back(objectName);
		++iter;
	}
	for (int i = 0; i < objects.size(); i++)
		if (strcmp(objects[i],"_GLOBAL") != 0)
			objectListWidget->addItem(QString(objects[i]));
	return objects;
} 


MainWindow::MainWindow(QWidget *parent): QMainWindow(parent)
{
	database = new BRLCAD::ConstDatabase();
	initMenu();
	setMinimumSize(800, 600);
}
